import sys
from typing import List, Set

from depcheck.models import DependencyReport, Ruleset


class DepChecker:
    __ruleset: Ruleset
    __dependency_report: DependencyReport

    def __init__(self, ruleset: Ruleset, dependency_report: DependencyReport) -> None:
        self.__ruleset = ruleset
        self.__dependency_report = dependency_report

    def run(self) -> None:
        # Check dependencies if there is any violation for given ruleset
        errors = {}

        for layer in self.__ruleset.rules:
            layer_defined_packages = self.__ruleset.layers[layer]
            detected_dependencies = self.__dependency_report.layer_dependencies(layer)

            print(f"\n### LAYER: '{layer}' WITH PACKAGES: {layer_defined_packages}")
            print(f"    DEPENDENCIES DETECTED:\n\t{detected_dependencies if detected_dependencies else 'none'}")

            violations = self.__check_rules(layer)
            if len(violations) > 0:
                errors[layer] = violations

        print("\n")

        if errors:
            print(f"!!! {len(errors)} ILLEGAL DEPENDENCIES FOUND:")
            packages_to_layers = self.__ruleset.packages_to_layers
            for layer, illegal_dependency_pkg in errors.items():
                for dependency in illegal_dependency_pkg:
                    illegal_layer_name = "?"
                    if dependency in packages_to_layers:
                        illegal_layer_name = packages_to_layers[dependency]
                    print(f"- Layer '{layer}' depends on '{dependency}' (Layer: {illegal_layer_name})")
            sys.exit(1)
        else:
            print("OK! All package dependencies look good!")
            sys.exit(0)

    def __check_rules(self, layer: str) -> List[str]:
        violation: List[str] = []
        layer_deps: Set[str] = self.__dependency_report.layer_dependencies(layer)
        whitelist: Set[str] = self.__dependency_report.whitelist(layer)

        black_list = layer_deps.difference(whitelist)

        if len(black_list) > 0:
            violation.extend(list(black_list))

        return violation



import json
import subprocess
from typing import Any, Dict

import yaml

from depcheck.models import DependencyReport, Ruleset


class RulesetReader:
    __filename: str

    def __init__(self, filename: str) -> None:
        self.__filename = filename

    def read(self) -> Ruleset:
        ruleset: Dict[str, Any]

        with open(self.__filename, "r") as stream:
            try:
                ruleset = yaml.safe_load(stream)
            except yaml.YAMLError:
                print(
                    "\n\n[!] Command must be run from project root " "and .depcheck.yml file should be in the root\n\n"
                )

        return Ruleset(ruleset["layers"], ruleset["whitelist"])


class DependencyReader:
    __ruleset: Ruleset
    __root_package: str

    def __init__(self, ruleset: Ruleset, root_package: str) -> None:
        self.__ruleset = ruleset
        self.__root_package = root_package

    def read(self) -> DependencyReport:
        output = (
            subprocess.check_output(
                f"pydeps --show-deps --no-show --no-output --max-bacon 2 {self.__root_package}", shell=True
            )
            .decode("u8")
            .strip()
        )

        dependency_graph = json.loads(output)

        return DependencyReport(self.__ruleset, dependency_graph)



from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

from depcheck.depchecker import DepChecker
from depcheck.models import DependencyReport, Ruleset
from depcheck.readers import DependencyReader, RulesetReader

DEFAULT_CONFIG_FILE_PATH = ".depcheck.yml"


def run_depcheck() -> None:
    command_args = create_argument_parser().parse_args()
    arg_config_file = command_args.file
    arg_root_package = command_args.root_package

    print(f"** Reading ruleset from '{arg_config_file}'")
    ruleset: Ruleset = RulesetReader(arg_config_file).read()
    print(f"** Reading project packages from root package: '{arg_root_package}'")
    dependency_report: DependencyReport = DependencyReader(ruleset, arg_root_package).read()
    print("** Checking dependencies")
    DepChecker(ruleset, dependency_report).run()


def create_argument_parser() -> ArgumentParser:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)

    parser.add_argument("root_package", help="Root package of the project")
    parser.add_argument(
        "-f",
        "--file",
        default=DEFAULT_CONFIG_FILE_PATH,
        help="Path to the config file that includes layers and rules.",
    )

    return parser


run_depcheck() if __name__ == "__main__" else None



from typing import List

from depcheck.util import flatten


def test_flatten_returns_items_of_inner_lists_in_a_flatten_list() -> None:
    # Arrange
    list_2d: List[List[str]] = [["a", "b", "c"], ["1", "2"], ["x", "y", "z"]]
    expected_flatten_list: List[str] = ["a", "b", "c", "1", "2", "x", "y", "z"]

    # Act
    result: List[str] = flatten(list_2d)

    # Assert
    assert expected_flatten_list == result



import sys
from copy import copy
from importlib import import_module
from typing import Any, cast

from depcheck.util import flatten


class Ruleset:
    __rules: dict[str, list[str]]
    __layers: dict[str, list[str]]

    def __init__(self, layers: dict[str, list[str]], rules: dict[str, list[str]]) -> None:
        self.__validate_layers(layers)
        self.__layers = layers
        self.__rules = rules

    @staticmethod
    def __validate_layers(layers: dict[str, list[str]]) -> None:
        for packages in layers.values():
            for package in packages:
                try:
                    import_module(package)
                except ModuleNotFoundError:
                    print(f"\tERROR: There is no package or module named '{package}'!")
                    sys.exit(1)

    def layer_packages(self, layer: str) -> list[str]:
        return self.layers.get(layer, [])

    def whitelist(self, layer: str) -> list[str]:
        return self.rules.get(layer, [])

    @property
    def rules(self) -> dict[str, list[str]]:
        return self.__rules

    @property
    def layers(self) -> dict[str, list[str]]:
        return self.__layers

    @property
    def packages_to_layers(self) -> dict[str, str]:
        out = {}
        for layer_name, packages_list in self.__layers.items():
            for package in packages_list:
                if package in out:
                    raise Exception(f"Package {package} is already in layer {out[package]}")
                out[package] = layer_name
        return out

    def packages_in_same_layer(self, layer: str, parent_package: str, detected_package: str) -> bool:
        if len(detected_package) < len(parent_package):
            # A package cannot be child of another package if the latter has shorter name
            return False

        if not detected_package.startswith(parent_package):
            return False

        pkg_to_layers = self.packages_to_layers
        if detected_package in pkg_to_layers and pkg_to_layers[detected_package] != layer:
            return False

        return True


class DependencyReport:
    __ruleset: Ruleset
    __graph: dict[str, Any]

    def __init__(self, ruleset: Ruleset, dependency_graph: dict[str, Any]) -> None:
        self.__ruleset = ruleset
        self.__graph = dependency_graph

    def layer_dependencies(self, layer: str) -> set[str]:
        # layer packages including subpackages detected
        layer_packages: set[str] = self.__layer_packages([layer])
        layer_dependencies = self.__package_dependencies(layer_packages)

        dependencies = self.__exclude_inner_dependencies(layer_dependencies, layer_packages)
        files = set()
        for dep in dependencies:
            maybe_module = import_module(dep)
            if not maybe_module:
                continue

            file_name = cast(str, maybe_module.__file__)
            if "__init__.py" not in file_name:
                files.add(dep)

        # Remove individual files and the root package
        return dependencies.difference(files.union({self.root_package}))

    def whitelist(self, layer: str) -> set[str]:
        whitelist: list[str] = self.__ruleset.whitelist(layer)

        return self.__layer_packages(whitelist)

    @property
    def root_package(self) -> str:
        if "__main__" not in self.__graph:
            print("\tERROR: Please make sure your root package has the `__init__.py` file and retry!")
            sys.exit(1)

        return self.__graph["__main__"]["imports"][0]

    @staticmethod
    def __exclude_inner_dependencies(raw_dependencies: set[str], layers: set[str]) -> set[str]:
        return raw_dependencies - layers

    def __layer_packages(self, layers: list[str]) -> set[str]:
        if layers is None:
            return set()

        # Packages within given layers
        layer_packages: list[str] = flatten([self.__ruleset.layer_packages(layer) for layer in layers])

        # Include subpackages of layer_packages
        for layer in layers:
            layer_packages.extend(self.__subpackages(layer, layer_packages))

        return set(layer_packages)

    def __subpackages(self, layer: str, parent_packages: list[str]) -> list[str]:
        subpackages: set[str] = set()

        detected_packages = list(self.__graph.keys())

        for parent_package in parent_packages:
            for detected_package in detected_packages:
                if self.__ruleset.packages_in_same_layer(layer, parent_package, detected_package):
                    subpackages.add(detected_package)

        pkg_to_layers = self.__ruleset.packages_to_layers

        subp_copy = copy(subpackages)
        for idx, sp in enumerate(subp_copy):
            for pkg, pkg_layer in pkg_to_layers.items():
                if sp.startswith(pkg) and pkg_layer != layer and sp in subpackages:
                    # Remove the subpackages that belong in different layers
                    # eg: if we have layer 1="a" and layer 2="a.b", then "a.b.c" does not belong in layer 1
                    subpackages.remove(sp)

        return list(subpackages)

    def __package_dependencies(self, packages: set[str]) -> set[str]:
        """
        Returns all packages that are used(imported) in given 'packages'.
        In other words, this returns dependencies of given 'packages'.
        """
        deps = set(flatten([self.__graph.get(i, {}).get("imports", []) for i in packages]))
        return set([item for item in deps if item not in packages])
